;; Programa: Gestor de ficheros de agentes para el juego PlanTron.
;;
;; Asignatura: Introducción a la Inteligencia Artificial
;;             ETSII - ULL
;; Práctica: (sistemas de producción)
;; Profesor: Patricio García (patricio AT etsii.ull.es)
;; Fecha:    23/04/2007
;; Comentarios:
;;           Las funciones aquí implementadas permiten gestionar
;;           los ficheros de agentes que pertenecen al directorio
;;           "arena". Controla también el fichero "hof.lisp" que
;;           contiene estadisticas de juego.
;;
;;           Además concluye cargando tres ficheros de agentes
;;           desde el directorio arena:
;;           - "pt-agentes-humanos" con agente-humano:
;;           agente activo para que un humano pueda jugar
;;           - "pt-agentes-simples": posee varios agentes
;;           activos sencillos
;;           - "pt-agentes-medios": posee varios agentes
;;           activos de nivel medio
;;
;;           Para poder utilizar estas funciones los agentes
;;           han de estar definidos bajo las normas descritas
;;           en el package pt (ver "pt.lisp" y ejemplos
;;           "pt-agentes-simples.lisp" y "agentes-aluXXXX.lisp")
;;
;;           En este código se define dos variables que indican
;;           el directorio y subdirectorio donde se encuentran
;;           los ficheros necesarios para cargar el entorno
;;           y el repositorio de agentes o 'arena'. Estas 
;;           variables son respectivamente PT:PATH y
;;           PT-ARENA:PATH. Estas variables, al cargarse, se
;;           inicializan al directorio donde está pt-loader y
;;           a el subdirectorio "arena" del directorio donde está
;;           PT-LOADER respectivamente (ver documentación
;;           PT:PATH y PT-ARENA:PATH).
;;
;;
;;           Las funciones accesibles mas importantes son:
;;
;;           LOAD-AG
;;           LS
;;           SEND
;;           DEL
;;           HOF
;;           STAT
;;
;;           Ver la documentación existente para cada una de ellas.
;;
;;           Para ver algunos ejemplos de llamadas a las fuciones
;;           aquí implementadas consultar "presentacion.lisp".
;;
;;
;;
;; Dependencias:
;;           Hace uso del paquete PT que contiene instrucciones
;;           que permiten un manejo cómodo y simplificado de
;;           agentes y partidas de PlanTron.
;;
;;           Hace uso de los paquetes PT-AGENTES-HUMANOS y
;;           PT-AGENTES-SIMPLES.
;;
;;
;; Modificaciones:
;;
;;           - Ver 0.6 (23/04/2007): Primera versión. Extraida de
;;             "pt-loader.lisp" a raiz de dividir este fichero en
;;             tres partes: "pt-loader.lisp", "plantron.lisp" y 
;;             "pt-arena.lisp".
;;
;;           - Ver 0.7 (1/05/2007): Ajustes para versión con nueva
;;             interface de comunicaciones en package "ENVIO"
;;
;;           - Ver 0.7.1 (15/05/2007): Incluido lis-lisp-types y 
;;             match-lisp-types de PT. Preparado send y del para
;;             SBCL. Arregaldo y dividido stat en hof y stat. Pasado
;;             la carga de agentes por defecto a pt-loader.
;;
;;           - Ver 0.7.2 (16/05/2007): Modificado match-lisp-types 
;;             para que no devuelva pathnames repetidos (problema
;;             en cmucl). Añadida función UNLOAD-AG.
;;
;;           - Ver 0.7.3 (13/04/2010): time-to-ymdhms y datos-* 
;;             pasado a pt.
;;
;;           - Ver 0.8 (28/04/2010): Implementado stat2
;;
;;
;; Por hacer:
;;           - Corrección de Bugs
;;             ------------------
;;           - Hacer que send-tron-arena y del-tron-arena funcionen
;;             para gcl.
;;           - send-tron-arena  canta /bin/tar: Member names contain `..
;;           - del-tron-arena canta "stty: entrada estándar: 
;;             Función ioctl no apropiada para el dispositivo"
;;
;;           - Mejoras en funcionalidades
;;             --------------------------
;;
;;           - Mejoras en estilo de programación
;;             ---------------------------------
;;

(unless (find-package "PT-ARENA")
  (load "pt-arena-package"))

(in-package "PT-ARENA")


; Variables version
; -----------------

(defconstant version "0.8" "Versión del package PT-ARENA.")

(defun version ()
  "Devuelve la versión del paquete PT-ARENA que se ha cargado."
  version)


;;
;; Variables a directorios y ficheros
;; ----------------------------------

(defvar path
  (if (directory (merge-pathnames (pathname "arena") pt:path))
      (merge-pathnames (pathname "arena/") pt:path)
    pt:path)
  "Directorio que contiene los distintos ficheros lisp de agentes jugadores de Pt.")

(defvar sender
  (merge-pathnames (pathname "sender-pt-arena") pt:path)
  "Programa encargado de enviar ficheros a arena.")

(defvar deleter
  (merge-pathnames (pathname "deleter-pt-arena") pt:path)
  "Programa encargado de eliminar ficheros de arena.")

(defvar lis-fich-load-arena nil
  "Lista de ficheros cargados en memoria del repositorio arena.")

(defvar lis-lisp-types '(".lisp")
  "Lista de tipos de ficheros LISP ordenada por preferencias.")

(defvar *hof-file* "hof.lisp"
  "Fichero que almacena del Hall of Fame de PlanTron")


;; Específico según implementación lisp
;; ------------------------------------

(defun pref-punt (st)
  (concatenate 'string "." st))

(let ((lit (lisp-implementation-type)))
  (cond
   ((string-equal lit "GNU Common Lisp (GCL)")
    (setf lis-lisp-types
	  (eval (find-symbol "*LOAD-TYPES*" "SYSTEM"))))

  ((string-equal lit "CMU Common Lisp")
   (use-package "EXTENSIONS")
   (setf lis-lisp-types
	 (mapcar #'pref-punt
		 (append (eval (find-symbol "*LOAD-OBJECT-TYPES*"))
			 (eval (find-symbol "*LOAD-SOURCE-TYPES*"))))))

   ((string-equal lit "CLISP")
    (setf lis-lisp-types
	 (mapcar #'pref-punt
		 (append (eval (find-symbol "*COMPILED-FILE-TYPES*"))
			 (eval (find-symbol "*SOURCE-FILE-TYPES*"))))))
   ((string-equal lit "SBCL")
    (setf lis-lisp-types
	  (mapcar #'pref-punt
		  (list (eval (find-symbol "*FASL-FILE-TYPE*" "SB-FASL"))
			(eval (find-symbol "*LOAD-SOURCE-DEFAULT-TYPE*" "SB-INT"))))))
   (T ; Desconocida
    (envio:my-warning "Implementación ~S no soportada." lit))))


;;
;; Funciones de caga de agentes
;; ----------------------------

(defun match-lisp-types (file path)
  "Busca ficheros en PATH que emparejen FILE o con alguna extensión de lisp."
  (delete-duplicates 
   (apply #'append
	  (mapcar #'(lambda (ext) 
		      (directory (merge-pathnames 
				  (pathname (concatenate 'string file ext)) 
				  path)))
		  (cons "" lis-lisp-types)))
   :test #'equal))


(defun load-ag (file-str)
  "Carga los agentes que se encuentren el los ficheros emparejados por FILE-STR dentro de repositorio arena. Para más información de las especificaciones de dichos ficheros ver PT:LOAD-AG."
  (let (load-files)
    ; Recorremos todos los ficheros pasados
    (dolist (file (match-lisp-types file-str path)
		  (reverse load-files))
      (if (member file lis-fich-load-arena :test #'equal)
	  (envio:my-warning "Fichero ~S ya fue previamente cargado." file)
	(progn
	  (if (pt:load-ag file)
	      (progn
		(push file load-files)
		(push file lis-fich-load-arena))))))))

(defun unload-ag (file-str)
  "Borra los agentes que se fueron cargados desdelos ficheros emparejados por FILE-STR dentro de repositorio arena mediante LOAD-AG."
  (let (unload-files)
    ; Recorremos todos los ficheros pasados
    (dolist (file (match-lisp-types file-str path)
		  (reverse unload-files))
      (if (not (member file lis-fich-load-arena :test #'equal))
	  (envio:my-warning "Fichero ~S no ha sido previamente cargado." file)
	(progn
	  (if (pt:unload-ag (pathname-name file))
	      (progn
		(push file unload-files)
		(setf lis-fich-load-arena
		      (delete file lis-fich-load-arena :test #'equal)))))))))


;;
;; Funciones de listado de agentes
;; -------------------------------

(defun ls (&optional (pat "*") full)
  "Devuelve los ficheros con agentes jugadores PlanTron almacenados actualmente en la sección arena. FULL si está a T devuelve un listado ampliado. PAT permite indicar un patron a emparejar."
  (let (lis-arena)
    (dolist (file (match-lisp-types pat path)
		  (reverse lis-arena))
      (if full
	  (push (pt:datos-file file) lis-arena)
	  (push (pathname-name file) lis-arena)))))


(defun datos-pkg (pkg)  ; Eliminar, no usado
  (let* ((pkg-name (package-name pkg))
	 (fich-name (first (member pkg-name lis-fich-load-arena 
				   :test (lambda (s p)
					   (string-equal s (pathname-name p)))))))
    (if fich-name
	(first (ls (pathname-name fich-name) t)))))


(defun nuevos (&optional agentes (hof *hof-file*))
  "Indica los agentes, de los cargados actaulmente en memoria, que no se encuentran en *HOF-FILE*. Si AGENTES esta instanciado busca solo con esa lista."
  (let* ((lis-ag (if agentes (pt::borra-ag-interactivo agentes) 
		   (pt::borra-ag-interactivo (pt:jug))))
	 (filename (merge-pathnames (pathname hof) pt:path))
	 (st (open filename :direction :input :if-does-not-exist :error))
	 (lis-res (read st)))
    (close st)
    (remove-if #'(lambda (ag)
		   (pt:busca-ag ag lis-res))
	       lis-ag)))


;;
;; Funciones de transferenica de agentes
;; -------------------------------------

(defun send (fich-agentes)
  "Envía un fichero de agentes PlanTron FICH-AGENTES (.lisp) al entorno arena."
  ;(declare (ignore fich-agentes))
  ;(and "Función aún no implementada" nil))
  (if (or (string-equal (lisp-implementation-type) "CMU Common Lisp")
	  (string-equal (lisp-implementation-type) "SBCL"))
      (if (directory sender)
	  (let ((exited
		 #+cmu  (run-program sender (list fich-agentes) :output t)
	         #+sbcl (sb-ext:run-program sender (list fich-agentes) :output t)
		 ))
	    (eq
	      #+cmu   (process-exit-code exited)
	      #+sbcl  (sb-ext:process-exit-code exited)
	     0))
	(envio:my-warning "No presente ~A." sender))
    (envio:my-warning "send-pt-arena no implementado para ~A, usar CMUCL o SBCL." (lisp-implementation-type))))


(defun del (fich-agentes)
  "Elimina el fichero de agentes PlanTron FICH-AGENTES (sin .lisp) del entorno arena. Ha de ser el propietario del fichero."
  ;(declare (ignore fich-agentes))
  ;(and "Función aún no implementada" nil))
  (if (or (string-equal (lisp-implementation-type) "CMU Common Lisp")
	  (string-equal (lisp-implementation-type) "SBCL"))
     (if (directory deleter)
	  (let ((exited 
		 #+cmu  (run-program deleter (list fich-agentes) :output t)
	         #+sbcl (sb-ext:run-program deleter (list fich-agentes) :output t)
		 ))
	    (eq
	     #+cmu   (process-exit-code exited)
	     #+sbcl  (sb-ext:process-exit-code exited)
	     0))
       (envio:my-warning "No presente ~A." deleter))
    (envio:my-warning "del-pt-arena no implementado para ~A, usar CMUCL o SBCL." (lisp-implementation-type))))


;;
;; Funciones estadísticas
;; ----------------------

(defun hof (&optional agentes (hof *hof-file*))
  "Muestra las estadisticas de partidas perdidas de agentes PlanTron de arena. Si se le pasa una lista AGENTES solo mostrará la lista con esos agentes. Las estadísticas serán extraidas del fichero HOF"
  (let* ((filename (merge-pathnames (pathname hof) pt:path))
	 (st (open filename :direction :input :if-does-not-exist :error))
	 (lis-res (read st)))
    (close st)
    (cond
     ((null agentes) lis-res)
     ((listp agentes)
      (mapcar #'(lambda (ag) (cons ag (pt:busca-ag ag lis-res)))
	      agentes))
     (t (cons agentes (pt:busca-ag agentes lis-res))))))


(defun stat-internal (args &rest pt-keys)
  "Genera las estadisticas de partidas perdidas de agentes PlanTron de arena. Si se le pasa una lista AGENTES entonces lanza a jugar (con PT:PARTIDA-TODOS) dichos agentes de AGENTES, si no utiliza los agentes no interactivos cargados en memoria. Devuelve y modifica el fichero HOF si es posible, se ha de tenese permiso de escritura sobre HOF."
  (let* ((agentes (first args))
	 (hof (or (second args) *hof-file*))
	 (filename (merge-pathnames (pathname hof) pt:path))
	 (st (open filename :direction :input :if-does-not-exist :error))
	 (lis-res (read st)))
    (close st)
    ; Modificar estadísticas añadiendo agentes
    (progn
      (setf lis-res
	    (apply #'pt::partida-todos-internal (list agentes lis-res) pt-keys))
      (setf st (ignore-errors (open filename :direction :output 
				    :if-does-not-exist :create
				    :if-exists :rename-and-delete)))
      (if st
	  (progn
	    (format st ";;;~%;;; Hall of fame de ~A~%;;; (no modificar manualmente)~%;;;~%" (envio:consulta-leyenda))
	    (print lis-res st)
	    (close st)))
      lis-res)))


(defun stat (&rest arglist)
  "Genera las estadisticas de partidas perdidas de agentes PlanTron de arena. Los parámetros que se le pasan realmente son (&OPTIONAL LIS-AG (HOF *HOF-FILE*) &KEY PT-KEYS). Si se le pasa una lista AGENTES entonces lanza a jugar (con PT:PARTIDA-TODOS) dichos agentes de AGENTES, si no utiliza los agentes no interactivos cargados en memoria. Devuelve y modifica el fichero HOF si es posible, se ha de tenese permiso de escritura sobre HOF."
  (multiple-value-bind (args lis-keys) 
      ; Se pone en ARGS los argumentos pasados
      ; y en LIS-KEYS la lista de keywords
      (pt::partition-args (cons :num pt::keywords-partida) arglist)
    (apply #'stat-internal args lis-keys)))


(defun stat2-internal (args &rest pt-keys)
  "Genera las estadisticas de partidas perdidas de dos en dos con agentes PlanTron de arena. Si se le pasan dos listas de LIS-AG1 y LIS-AG2 de agentes, entonces lanza a jugar (con PT:PARTIDA-TODOS2) dichos agentes, si no utiliza los agentes no interactivos cargados en memoria. Devuelve y modifica el fichero HOF si es posible, se ha de tenese permiso de escritura sobre HOF."
  (let* ((agentes1 (first args))
	 (agentes2 (second args))
	 (hof (or (third args) *hof-file*))
	 (filename (merge-pathnames (pathname hof) pt:path))
	 (st (open filename :direction :input :if-does-not-exist :error))
	 (lis-res (read st)))
    (close st)
    ; Modificar estadísticas añadiendo agentes
    (progn
      (setf lis-res
	    (apply #'pt::partida-todos2-internal (list agentes1 agentes2 lis-res) pt-keys))
      (setf st (ignore-errors (open filename :direction :output 
				    :if-does-not-exist :create
				    :if-exists :rename-and-delete)))
      (if st
	  (progn
	    (format st ";;;~%;;; Hall of fame de ~A~%;;; (no modificar manualmente)~%;;;~%" (envio:consulta-leyenda))
	    (print lis-res st)
	    (close st)))
      lis-res)))


(defun stat2 (&rest arglist)
  "Genera las estadisticas de partidas perdidas de dos en dos con agentes PlanTron de arena. Los parámetros que se le pasan realmente son (&OPTIONAL LIS-AG1 LIS-AG2 (HOF *HOF-FILE*) &KEY PT-KEYS). Si se le pasan dos listas de LIS-AG1 y LIS-AG2 de agentes, entonces lanza a jugar (con PT:PARTIDA-TODOS2) dichos agentes, si no utiliza los agentes no interactivos cargados en memoria. Devuelve y modifica el fichero HOF si es posible, se ha de tenese permiso de escritura sobre HOF."
  (multiple-value-bind (args lis-keys) 
      ; Se pone en ARGS los argumentos pasados
      ; y en LIS-KEYS la lista de keywords
      (pt::partition-args (cons :num pt::keywords-partida) arglist)
    (apply #'stat2-internal args lis-keys)))

