;; Programa: Gestor de agentes y partidas de juego PlanTron.
;;
;; Asignatura: Introducción a la Inteligencia Artificial
;;             ETSII - ULL
;; Práctica: (sistemas de producción)
;; Profesor: Patricio García (patricio AT etsii.ull.es)
;; Fecha:    23/04/2007
;; Comentarios:
;;           Las funciones aquí implementadas permiten manejar
;;           de una manera cómoda los agentes utilizables
;;           para el entorno de juego PlanTron (ver pt-base y
;;           envio). Facilitando la carga en memoria de
;;           todo el entorno y los agentes y consulta de los 
;;           agentes cargados en memoria y sus características.
;;
;;           Para poder utilizar estas funciones los agentes
;;           han de estar definidos bajo las siguientes normas
;;           (ver ejemplo "pt-agentes-humanos.lisp" y
;;           "pt-agentes-simples.lisp"):
;;
;;           - Se ha de crear un fichero que contenga la
;;           definición de la/s funcion/es de el/los agentes
;;           que se deseen crear.
;;           - El fichero debe de contener definir un package
;;           cuyo nombre ha de conincidir con el nombre del
;;           fichero sin extensión. Dicho package ha de usar
;;           el entorno de comunicaciones "ENVIO" y ha de
;;           exportar todas las funciones que definen cada
;;           agente.
;;           - Además el package anterior debe exportar una
;;           constante denominada 'agente', dicha constante
;;           ha de iniciallizarse en el fichero (defconstant)
;;           a una lista con una sublista por cada agente
;;           que se quiera definir. Dichas sublistas contendrán
;;           las funciones que definen cada agente (ver
;;           documentación de variable LIS-AGENTES y
;;           documentación de envio).
;;
;;           En este código se define una variables path que 
;;           indica el directorio con los ficheros necesarios
;;           para cargar el entorno Pt. Al cargarse, se
;;           inicializan al directorio donde está pt-loader.
;;
;;           Este entorno carga automáticamente los ficheros
;;           con los siguientes agentes:
;;           - "pt-vis" con agente-vis: agente pasivo
;;           encargado de la visulización de partidas
;;
;;           Los símbolos accesibles mas importantes son:
;;
;;           LOAD-AG
;;           JUG
;;           INFO-JUG
;;           PARTIDA
;;           PARTIDA-TODOS
;;           VERSION
;;           *REP*
;;           *VIS*
;;           *GEN-PLA*
;;
;;           Ver la documentación existente para cada una de ellos.
;;
;;           A continuación se muestran algunos ejemplos de llamadas
;;           a las fuciones aquí implementadas:
;;
;;
;;           - Consulta de agentes en memoria:
;;             (pt:jug)
;;
;;           - Iniciar partida por defecto, agente humano:
;;             (pt:partida)
;;
;;           - Iniciar partida entre dos agentes dados:
;;             (pt:partida "AGENTE-ESPIRAL" "AGENTE-HORMIGA-BORRACHA")
;;
;;           - Carga de fichero de agentes "agentes-aluXXX.lisp":
;;             (pt:load "agentes-aluXXX.lisp")
;;
;;           - Consulta 'full' de agentes en memoria:
;;             (pt:jug :full t)
;;
;;           - Consulta de información de agentes en memoria:
;;             (pt:info-jug "AGENTE-ESPIRAL")
;;
;;          - Jugar todos agentes entre sí rápidamente
;;            y sin visualización:
;;             (pt:partida-todos :sleep-cs 0 :vis nil)
;;
;;
;; Dependencias:
;;           Hace uso del paquete que contiene la interface (ENVIO)
;;           con el servidor del juego (PT-BASE).
;;
;;           Hace uso del paquete PT-VIS.
;;
;;           Hace uso del paquete PT-AGENTES-GENPLA
;;
;;
;; Modificaciones:
;;
;;           - Ver 0.1 (28/04/2006): Primera versión.
;;
;;           - Ver 0.2 (02/05/2006): Incluido tron-todos.
;;
;;           - Ver 0.3 (03/05/2006): Carga automáticamente variable
;;             path-tron desde *load-truename*. Implementada 
;;             lis-fich-load-arena para controlar ficheros cargados
;;             del arena.
;;
;;           - Ver 0.3.1 (03/05/2006): Corregido bug que confundía
;;             el identificador del jugador enviado a funciones de
;;             inicialización de agentes (no enviaba nil, agente
;;             humano como jugador 2 confundía teclas). Corregido
;;             bug para cargar con GCL (*load-patname* y
;;             *load-trueanem*).
;;
;;           - Ver 0.3.2 (04/05/2006): Ampliada key :vis de tron
;;             para contemplar paso de string y abrirlo como stream
;;             de salida (*tron-vis-stream*) para tron-vis,
;;             ver v. 0.2 de tron-vis-xterm.
;;
;;           - Ver 0.3.3 (05/05/2006): Añadiad función version-tron.
;;             Ampliada opción full de jug-tron. Utilizado
;;             merge-pathnames para unir directorios y ficheros.
;;             Corregido bug de cargar varias veces tron-vis en gcl. 
;;
;;           - Ver 0.3.4 (11/05/2006): Implementadas funciones 
;;             send-tron-arena y del-tron-arena.
;;
;;           - Ver 0.4 (12/05/2006): Cambios en tron-internal debido
;;             a cambio en tron-comm de ver "fun-0.3". Tratamiento
;;             a distinto tipo de visualizadores (:vis) y desplazado
;;             dispositivo de visualización a :dev-vis.
;;
;;           - Ver 0.4.1 (18/05/2006): Cambiado send-tron-arena y
;;             del-tron-arena para llamar a programs externos
;;             almacenados en las variables sender-tron-arena y
;;             deleter-tron-arena respectivamente.
;;
;;           - Ver 0.4.2 (20/05/2006): Modificacines sender y
;;             del-tron-arena.
;;
;;           - Ver 0.4.3 (27/05/2006): Modificacin tron-todos para
;;             mostrar número de partidas jugadas. Añadido función
;;             stat-tron-arena con actualizazión del "hof.lsp".
;;
;;           - Ver 0.4.4 (29/05/2006): Añadido nuevos-tron-arena .
;;
;;           - Ver 0.5 (30/05/2006): Carga paquetes shell, q_search
;;             y a_star. Añadido :sleep-nil en tipo-juego.
;;
;;           - Ver 0.5.1 (02/06/2006): Añadido opción hof de
;;             stat-tron-arena y nuevos-tron-arena.
;;
;;           - Ver 0.5.3 (30/06/2006): Añadido opciones file y usr
;;             a jug-tron.
;;
;;           - Ver 0.5.4 (17/07/2006): Corregido error en 
;;             stat-tron-arena
;;
;;           - Ver 0.6 (21/4/2007): Cambio de nombre de tron-loader a 
;;             pt-loader. Cambio extensión de "lsp" a "lisp". Cambio 
;;             también nombre de variables, funciones, etc de "tron" a 
;;             "pt". División en tres de "pt-loader": "pt-loader.lisp",
;;             "plantron.lisp" y "pt-arena.lisp".
;;
;;           - Ver 0.7 (1/05/2007): Ajustes para versión con nueva
;;             interface de comunicaciones en package "ENVIO"
;;
;;           - Ver 0.7.1 (2/05/2007): Incorporada funcion INFO-JUG.
;;             Ajustes en documentación. Incorporado key :num-jug-pas
;;             al tipo-partida. Reparado problema con conteo de
;;             agentes gpla y pasivos falla
;;
;;           - Ver 0.7.2 (8/05/2007): Reparación de bug que congelaba
;;             ejecución de partida con agente inexistente. Retoques
;;             en documentación.
;;
;;           - Ver 0.7.3 (11/05/2007): Modificacipn de partida-todos:
;;             inclusión de parámetro LIS-RES y key :num.
;;
;;           - Ver 0.7.4 (17/05/2007): Pasado lis-lisp-types y 
;;             match-lisp-types a PT-ARENA. Añadida funciones
;;             DEL-AG y UNLOAD-AG. Modificado partida-todos-internal
;;             con lista de resultados a ampliar y opción :num.
;;             Parecheado jug para que funcione full
;;
;;           - Ver 0.7.5 (22/05/2007): Añadida carga de
;;             "pt-util.lisp". Implementada opción :use
;;
;;           - Ver 0.7.6 (24/05/2007): Modificado obtiene-*-ag para
;;             permitir trabajar con listas en datos de keywords
;;
;;           - Ver 0.7.7 (22/04/2008): Corregido error de nombre de
;;             *GENPLA* y keyword genpla a *GEN-PLA* y gen-pla.
;;             Añadido keywords :pla-car y :mov-con (ver 1.2 de 
;;             pt-base).
;;
;;           - Ver 0.8 (26/04/2008): Modificado para cargar
;;             pt-agentes-genplc en vez de pg-agentes-genpla
;;
;;           - Ver 0.9 (03/04/2009): Inclusión de :mov-con, :tam-col,
;;             :for-cab y :for-col. Uso de defagente. Uso de :gen-dir
;;
;;           - Ver 0.9.2 (22/04/2009): Inicio por defecto en agentes
;;             en posiciones aleatorias. Opción :tip-col para jug.
;;             Incorporada carga de agentes depredadores dpdd-minmax
;;
;;           - Ver 0.9.3 (08/05/2009): Modificado 
;;             partida-todos-internal para ordenar por pasos mas
;;             energía acumulada final. Corregido el error en calculo
;;             de resultados de partida-todos-internal
;;
;;           - Ver 0.9.4 (23/03/2010): Adaptaciones para tron clásico:
;;             desactivamos generadores de plantas y direcciones por
;;             defecto, con movimiento continuo y cola. Modificado
;;             bindkey para aceptar keydef. Juega con 2 humanos por
;;             defecto.
;;
;;           - Ver 0.9.5 (12/04/2010): Incluida inicialización de
;;             pt-base::*path-maps*. Opción :tip-tab y corregido
;;             problema por falta de :pos-alea en keywords-partida.
;;             Incluído keyword :file para describir fichero de
;;             procedencia de agente eliminando llamada a
;;             pt-arena::datos-pkg desde jug.
;;
;;           - Ver 0.9.6 (23/03/2011): Adaptación para Depredador/
;;             Presa con plantas comestibles
;;
;;
;; Por hacer:
;;           - Corrección de Bugs
;;             ------------------
;;           - Chequear funcionamiento en clisp:
;;             (directory (concatenate 'string path-tron-arena ""))
;;             con resultado distinto.
;;           - Versiones que no signan los links para clisp y gcl
;;             de load-tron.
;;           - Chekear conflictos al cargar agentes con igual nombre.
;;           - directory (en load-tron) expande los nombres de
;;             ficheros añadiendo * al final. Limitarlos con
;;             lis-lisp-types.
;;           - En datos-file cmucl 19a salta error al ejecutar
;;             file-autchor y file-write-date por segunda vez
;;           - Con unload-ag pueden darse problemas: si al cargar
;;             un fichero de agentes alguno de los agentes tiene
;;             el mismo nombre que uno presente en memoria y no
;;             se carga, al borrar ese package borra el agente
;;             de otro.
;;           - ¿Utilización de :use con lista con agente no pasivo?
;;           - ¿:interactivo debería estar tambien definido en
;;             keywords-nofun-ag de pt-base? variableskeywords-info-ag
;;             y keywords-ag no se usan
;;           - obtiene-nombre-ag ya no tiene sentido, borrar
;;           - En inserta-ag caga erroneo el campo :pkg
;;
;;           - Mejoras en funcionalidades
;;             --------------------------
;;           - Chequear que pasa si no se ha definido varialble
;;             con lista de agentes del fichero en load-tron.
;;           - Añadir agente log de partidas.
;;           - Guardar info que relacione agente - fichero.
;;           - ¿Función que descarge agentes de memoria (unload)?
;;             (delete-package).
;;           - Permitir más de un generador de plantas
;;           - ¿Qué inserta-ag añada también referencia a package
;;             y fichero de donde se extrajo el agente?
;;           - Que no sea necesario pasar los agentes ordenados
;;             por gen-dir, gen-pla y pasivos
;;           - Que acepten Ctrl-L para refrescar pantalla si no
;;             son agentes interactivos
;;           - Codificar unload-ag
;;           - Chequear keywords de agentes válidas con keywords-ag
;;           - Cambiar estadísticas de partida-todos2 para que
;;             refleje la energía final de los agentes
;;
;;           - Mejoras en estilo de programación
;;             ---------------------------------
;;           - Añadir especificaciones parámetros a documentación.
;;           - Sustituir bind-key por getf
;;
;;

(unless (find-package "PT")
  (load "pt-package"))

(in-package "PT")


;; Variables version
;; -----------------

(defconstant version "0.9.6" "Versión del package PT.")

(defun version ()
  "Devuelve la versión del paquete PT que se ha cargado."
  version)


;;
;; Variables globales
;; ------------------

(defvar path
  (merge-pathnames
   (directory-namestring
    (eval (or (find-symbol "*LOAD-TRUENAME*")
	      (find-symbol "*LOAD-PATHNAME*")
	      (find-symbol "*LOAD-PATHNAME*" "SYSTEM") ; Para GCL
	      ""))))
  "Directorio que contiene los distintos ficheros necesarios para el juego Pt.")

(defvar lis-agentes nil
  "Lista asociativa de los agentes cargados en memoria actualmente.")
;; Contiene listas con un elemento para cada agente cagado en memoria
;; que incluye:
;;         FIRST: nombre del agente (string campo clave)
;;         REST: lista de keywords del agente
;;               ver descripción en keywords-ag
;;
;; Estos datos son extraidos a partir del contenido de las keywords
;; de la definición defagente
;;
(defconstant keywords-info-ag
  '(:name :documentation :tipo-ag :mov-con :tam-col :for-cab :for-col :interactivo :use)
  "Lista de posibles keywords relacionada con información de los agentes admitidas en el contenido del símbolo AGENTES que debe estar presente en cada package de agentes de la definición de los agentes (ver PT:LOAD-AG). El significado de cada una de las keywords admisibles son:
  :NAME          : Nombre del agente
  :PKG           : Pakete en que esta integrado el agente
  :DOCUMENTATION : Documentación sobre el agente
  :TIPO-AG       : Tipo de agente, puede tomar uno alguno de los valores
                   presentes en PT-BASE:TIPOS-AG. En caso de que no exista toma
                   el valor ENVIO:GEN-DIR
  :MOV-CON       : Si que el agente debe tener movimiento contínuo, esto es,
                   (en última dirección elegida) (T) o si se puede detener
                   (NIL). En caso de que no exista toma el valor de 
                   PT-BASE::MOV-CON. Solo apto para agentes tipo gen-dir
  :TAM-COL       : Si que el agente debe dejar rastro al moverse,
                   (con cola) (T) o si no debe dejar rastro (NIL)
                   En caso de que no exista toma el valor
                   de PT-BASE::TAM-COL. Solo apto para agentes tipo gen-dir
  :FOR-CAB       : Valor numérico que indica la fortaleza de la cabecera del
                   agente (ver ...). En caso de que no exista toma el valor
                   de pt-base::for-cab
  :FOR-COL       : Valor numérico que indica la fortaleza de la cola del
                   agente (ver ...). En caso de que no exista toma el valor
                   de pt-base::for-col
  :INTERACTIVO   : Si su valor es T indica que el agente necesita interactuar
                   con un humano, como el caso del AGENTE-HUMANO. Útil para
                   evitar enfrentamientos en background (ver PT:PARTIDA-TODOS)
  :USE           : Lista con nombre de agentes pasivos que usa el agente que se
                   define. Han de estar cargados en memoria y se añadirán como
                   agentes pasivos a la partida
")

(defconstant keywords-ag
  (append keywords-info-ag envio:keywords-fun-ag)
  "Lista de posibles keywords admitidas en el contenido del símbolo AGENTES que debe estar presente en cada package de agentes de la definición de los agentes. Ver descripción de KEYWORDS-INFO-AG Y ENVIO:KEYWORDS-FUN-AG.")

(defconstant keywords-partida
  '(:rep :vis :gen-pla :gen-dir :mov-con :for-cab :for-col :dev-vis :num-jug :tip-tab :dim-tab :sin-bordes :sleep-cs :no-verbose :asincrono :sleep-nil :pos-alea)
  "Lista de posibles keywords admitidas en la llamada a la función PARTIDA")

(defvar lis-pkg-ag-load nil
  "Lista de packages de agentes cargados en memoria.")

(defvar lis-ag-vis nil
  "Lista de funciones de agentes visualizadores.")

(defvar lis-ag-genpla nil
  "Lista de funciones de agentes generadores de plantas.")

(defvar lis-ag-gendir nil
  "Lista de funciones de agentes generadores de direcciones para :gen-dir.")

(defvar *lis-nom-agentes-def* '("HUMANO")
  "Nombre del agente que juega por defecto si no se le pasa ninguno")

(defvar *rep* 1
  "Indica el número de réplicas que se realizarán de los agentes pasados para ejecutar una partida del juego PlantTron.")

(defvar *vis* 1
  "Indica el número de agente visualizador a utilizar en partidas del juego PlanTron. NIL significa no utilizar agente visualizador.")

(defvar *dev-vis* t
  "Indica el nombre del dispositivo a utilizar por agente visualizador en partidas del juego PlanTron para su salida, normalmente T ó algo como \"/dev/pts/14\".")

(defvar *log* nil
  "Indica si utilizar agente log en partidas del juego PlanTron.")

(defvar *gen-pla* 1
  "Indica cual agente generador de plantas utilizar fijo en partidas del juego Pt. NIL no usa agente generador de plantas")

(defvar *gen-dir* 0
  "Indica cual agente generador de direcciones utilizar fijo en partidas del juego Pt. NIL no usa agente generador de direcciones")

(defvar id-ag-eval nil
  "Identificador del agente evaluado")

(defvar file-loading nil
  "Información del fichero que se está cargando")


;;
;; Funciones para trabajo con espec. de agentes (lista de keywords)
;; --------------------------------------------

(defun bind-key (key lis-keys &optional keydef)
  "Obtiene el valor asociado a la key KEY de la lista de keys LIS-KEY si la encuentra y si no devuelve KEYDEF."
  (getf lis-keys key keydef))

(defun del-key (key lis-keys)
  "Elimina la key KEY y su valor asiciado de la lista de key LIS-KEY."
  (remf lis-keys key))


;;
;; Funciones de caga de agentes
;; ----------------------------

(defun busca-ag (ag &optional (lista lis-agentes))
  "Busca agente AG en la lista asociativa LISTA y devuelve su lista de keys."
  (rest (assoc ag lista :test #'string=)))

(defun inserta-ag (pkg nombre &rest lis-keys)
  "Inserta agente definido por la lista de keys LIS-KEY en la lista asociativa LIS-AGENTES siempre que no exista ya uno con igual nombre."
  (let ((doc nil))
    (if (stringp (first lis-keys)) ; Si se inclye documentación, añadirla
	(setf doc (first lis-keys)
	      lis-keys (rest lis-keys)))
    ; Ver si cambiar nombre por tag :name
    (if (getf lis-keys :name)
	(setf nombre (getf lis-keys :name))
      (setf (getf lis-keys :name) nombre))
    ; Asegurar que no exista un agente en memoria con igual nombre
    (if (busca-ag nombre)
	(envio:my-warning "Ya existe un agente con el nombre ~S en memoria, si desea actualizarlo descargue previamente todo su package (función UNLOAD-AG)." nombre)
      (progn
	; Inserta, si no existe, keywords :pgk, :documentation, :tipo-ag y :file
	(setf (getf lis-keys :pkg) pkg)
	(if (and (null (getf lis-keys :documentation)) doc)
	    (setf (getf lis-keys :documentation) doc))
	(if (null (getf lis-keys :tipo-ag))
	    (setf (getf lis-keys :tipo-ag) 'envio:gen-dir))
	(setf (getf lis-keys :file) file-loading)
	; Insertar nuevo agente
	(push (cons nombre lis-keys) lis-agentes)
	t))))

(defmacro defagente (nombre &rest lis-keys)
  "Definición de agente con nombre NOMBRE y según lista de keys LIS-KEY que puede contener las keys contenidas en KEYWORDS-AG"
  (let ((sn (symbol-name nombre)))
    (append `(inserta-ag *package* ,sn) lis-keys)))

(defun load-ag-1 (file)
  "Carga y devuelve los agentes que se encuentren en el fichero de pathname FILE."
  (if (not (load file :verbose t :if-does-not-exist nil))
      (envio:my-warning "Fichero ~S no se pudo cargar." file)
    (let ((pkg (find-package (string-upcase (pathname-name file)))))
      (if (null pkg)
	  (envio:my-warning "Fichero ~S no parece ser tipo agente PlanTron (no encontrado PACKAGE)." file)
	(push pkg lis-pkg-ag-load)))))

(defun time-to-ymdhms (time)
  "Transforma de universal-time a lista de año mes dia hora minuto y segundo."
  (multiple-value-bind (s m h d mt y)
      (decode-universal-time time)
    (list y mt d h m s)))

(defun datos-file (file)
  "Etrae lista de datos de un fichero FILE"
  (list (pathname-name file)
	#-cmu (file-author file)  ; FixMe
	#-cmu (time-to-ymdhms (file-write-date file))  ; FixMe
	))

(defun load-ag (file-str)
  "Carga los agentes que se encuentren definidos en los ficheros emparejados por FILE-STR. Los ficheros que definen agentes han de contener la especificación de un package con igual nombre que el fichero sin extensión. En dicho package ha de existir una constante o variable con nombre AGENTES cuyo contendido es una lista con una de listas de keywords por cada agente definido. Las keywords admisibles son las definidas en PT-BASE:KEYWORDS-FUN-AG que definen las funciones que conforman un agentes más las definidas en PT::KEYWORDS-INFO-AG que definen información relativa al agente."
  (let (load-files)
    ; Recorremos todos los ficheros pasados
    (dolist (file (directory file-str) ; :truenamep nil)) ;en CMUCL
		  (reverse load-files))
      (setf file-loading (datos-file file))
      (if (load-ag-1 file)
	  (push file load-files))
      (setf file-loading nil))))

(defmacro push-no-nil (elm lis)
  `(let ((elemento ,elm))
     (if elemento
	 (push elemento ,lis))
     elemento))

(defmacro del-ag-m (lis-ag &optional (lis 'lis-agentes))
  "Borra la lista de agentes LIS-AG de las lista LIS-AGENTES."
  `(let (del-lis)
     (setf ,lis (delete-if #'(lambda (elm)
			       (push-no-nil
				(first (member (first elm)
					       ,lis-ag
					       :test #'string=))
				del-lis))
			   ,lis))
     del-lis))

(defun del-ag (ags)
  "Borra el agente AGS o la lista de agentes AGS de las lista de agentes jugadores."
  (let ((lis-ag (if (consp ags) ags (list ags))))
    (del-ag-m lis-ag)))

(defun unload-ag (pkg-name) ; No usado, eliminar
  "Borra el package de agentes PKG-NAME y sus agentes asociados que hayan sido cargados mediante LOAD-AG."
  (let ((pkg (find-package (string-upcase pkg-name))))
    (if pkg
	(let ((var-ag (find-symbol "AGENTES" pkg)))
	  (if var-ag
	      (let ((lis-ag-var-ag (mapcar #'obtiene-nombre-ag (eval var-ag))))
		(setf lis-pkg-ag-load (delete var-ag lis-pkg-ag-load))
		(del-ag-m lis-ag-var-ag lis-agentes)
		(del-ag-m lis-ag-var-ag lis-ag-vis)
		(del-ag-m lis-ag-var-ag lis-ag-genpla)))
	  (delete-package pkg)))))

(defun unload-ag (pkg-name) ; Codificar
  (declare (ignore pkg-name))
  "Borra el package de agentes PKG-NAME y sus agentes asociados que hayan sido cargados mediante LOAD-AG."
  "Función aún no implementada, salir de lisp y volver a cargar el entorno")

;;
;; Funciones de listado de agentes
;; -------------------------------

(defun borra-ag-interactivo (lis-ag)
  "Elimina agente interactivo (como el AGENTE-HUMANO) de la lista LIS-AG."
  (remove-if #'(lambda (ag)
	       (bind-key :interactivo (busca-ag ag)))
	     lis-ag))

(defun info-jug (agente)
  "Devuelve la información del jugador AGENTE pasado como una lista de propiedades."
  (busca-ag agente))
  

(defun jug (&key full usr file (tipo-ag  'envio:gen-dir) (tam-col pt-base:*tam-col*))
  "Devuelve los agentes jugadores tipo generadores de direcciones cargados actualmente en memoria. FULL si está a T devuelve un listado ampliado. USR y FILE corresponden a información del usuario y fichero si se quiere filtrar el listado. TIPO-AG indica el tipo de agente a listar, si está a NIL listar todos. TAM-COL indica el tipo de cola que tiene el agente a listar, si está NIL los sin cola y T con cola."
  (let (lis-fun-ag)
    (dolist (ag lis-agentes lis-fun-ag)
      (if (and (or (not tipo-ag) (eq (bind-key :tipo-ag (rest ag)) tipo-ag))
	       (eq (bind-key :tam-col (rest ag) pt-base:*tam-col*) tam-col))
	  (if full
	      (let ((dat (bind-key :file (rest ag))))
		(if (and (or (not file) (string= file (first dat)))
			 (or (not usr) (string= usr (second dat))))
		    (push (list (first ag)
				dat
				(bind-key :documentation (rest ag)))
			  lis-fun-ag)))
	    (push (first ag) lis-fun-ag))))))


;;
;; Funciones para lanzar partidas
;; ------------------------------

(defun partida-internal2 (lis-jug tipo-partida)
  "Inicia partida PlanTron entre agentes pasados en LIS-JUG. TIPO-PARTIDA contiene el resto de keys que definen el tipo juego de la partida."
  (let (lis-lis-key-ag lis-pasivos lis-lis-key-ag-pasivos)

    ; Generamos lista de keys de agentes de jugadores pasados
    ; y revisamos que esten cargados agentes
    (dolist (ag lis-jug)
      (let ((lis-ag (busca-ag ag)))
	(if (null lis-ag)
	    (error "Agente ~S no está cargado en memoria." ag)
	  (progn
	    (push lis-ag lis-lis-key-ag)
	    (setf lis-pasivos (append (reverse (getf lis-ag :use)) lis-pasivos))))))

    ; Generamos lista de keys de agentes pasivos indirectos
    ; y revisamos que esten cargados
    (dolist (ag (delete-duplicates lis-pasivos :test #'string=))
      (let ((lis-ag (busca-ag ag)))
	(if (null lis-ag)
	    (error "Agente ~S no está cargado en memoria." ag)
	  (push lis-ag lis-lis-key-ag-pasivos))))

    (setf lis-lis-key-ag (nconc lis-lis-key-ag lis-lis-key-ag-pasivos))
 
    ; Contar número de jugadores gendir, genpla y pasivos
    (let ((num-jug-gdir 0)
	  (num-jug-gpla 0)
	  (num-jug-pas  0))
      (dolist (lis-key-ag lis-lis-key-ag)
	(let ((tipo (bind-key :tipo-ag lis-key-ag)))
	  (case tipo
	    (envio:gen-dir (setf num-jug-gdir (1+ num-jug-gdir)))
	    (envio:gen-pla (setf num-jug-gpla (1+ num-jug-gpla)))
	    (envio:pasivo  (setf num-jug-pas  (1+ num-jug-pas)))
	    (t (envio:my-warning "Tipo de agente ~S desconocido" tipo)))))

      ; Actualizar datos de número de agentes en tipo partida
      (setf (getf tipo-partida :num-jug-gdir) num-jug-gdir)
      (setf (getf tipo-partida :num-jug-act)  (+ num-jug-gdir num-jug-gpla))
      (setf (getf tipo-partida :num-jug-pas)  num-jug-pas))

     ; Enviamos petición de inicio de partida
    (envio:pet-ini lis-lis-key-ag
		   tipo-partida)))

 
(defun partida-internal (lis-nom-agentes
			 &rest
			 lis-keys 
			 &key
                         (rep *rep*)
			 (vis *vis*) 
                         ; (log nil)
			 (gen-pla *gen-pla*)
			 (gen-dir *gen-dir*)
			 (dev-vis *dev-vis*)
			 &allow-other-keys)
    "Inicia partida PlanTron entre agentes pasados en LIS-NOM-AGENTES. La key REP indica en numero de replicas que se generarán de los agentes pasados.VIS indica el número de agente-pt-vis que se  utilizará (1) o si no se utiliza (nil). GEN-PLA indica el número o lista de numeros de agentes generadores de plantas a utilizar. GEN-DIR indica el número o nombre, o lista de numeros o nombres de agentes generadores de direcciones fijos a utilizar. DEV-VIS indica un dispositivo normalmente de terminal (/dev/pts/*) donde emitir la salida del agente visualizador seleccionado. El resto de keys definen el tipo de partida a jugar (ver PT-BASE::EXTRC-TIPO-PARTIDA)."

    ; Agentes por defecto
    (if (null lis-nom-agentes)
	; Por defecto juega contenido *lis-nom-agentes-def*
	(setf lis-nom-agentes *lis-nom-agentes-def*))
    (setf id-ag-eval 1)

    ; Tratar las réplicas
    (let ((l lis-nom-agentes))
      (dotimes (i (1- rep))
        (setf lis-nom-agentes (append l lis-nom-agentes))))

    ; Ver si añadir a la lista de agentes otros agentes extras
    (if (not (listp gen-dir))
	(setf gen-dir (list gen-dir)))
    (if (consp gen-dir) ; Utilizar agente(s) generador(es) de direcciones
	(dolist (num-ag gen-dir)
	  (let ((nom-gendir (if (numberp num-ag)
				(first (nth num-ag lis-ag-gendir))
			      num-ag)))
	    (if (null nom-gendir)
		(error "Número de agente generador de direcciones (:gendir) no adecuado.")
	      ; Añadir agente(s) generador(es) de direcciones
	      (push nom-gendir lis-nom-agentes)))))
    (setf id-ag-eval (+ id-ag-eval (length gen-dir)))

    ; Dar la vuelta para mantener el orden con push
    (setf lis-nom-agentes (reverse lis-nom-agentes))

    ; Ver si añadir a la lista de agentes otros agentes extras
    (if (numberp gen-pla)
	(setf gen-pla (list gen-pla)))
    (if (consp gen-pla) ; Utilizar agente(s) generador(es) de plantas
	(dolist (num-ag gen-pla)
	  (let ((nom-genpla (first (nth num-ag lis-ag-genpla))))
	    (if (null nom-genpla)
		(error "Número de agente generador de plantas (:genpla) no adecuado.")
	      ; Añadir agente(s) generador(es) de plantas
	      (push nom-genpla lis-nom-agentes)))))

    (if (numberp vis)  ; Utilizar un agente de visualización
	(let ((nom-vis (first (nth vis lis-ag-vis))))
	  (if (null nom-vis)
	      (error "Número de agente visualizador no adecuado.")
	    ; Añadimos un agente visualizador
	    (progn
	      (push nom-vis lis-nom-agentes)
	      ; Abrir streamer para pt-vis, FIXME de otra manera
	      (let ((st-vis (find-symbol "*PT-VIS-STREAM*" "PT-VIS")))
		(if st-vis
		    (set st-vis
			 (cond
			  ((eq dev-vis t) *standard-output*)
			  ((eq dev-vis nil) nil)
			  (t
			   (open dev-vis :direction :output
				 :if-exists :overwrite))))))))))

    ; Añadimos el agente log, FIXME implementarlo primero

    ; Eliminar key vis, genpla y dev-vis FIXME
    (del-key :rep lis-keys)
    (del-key :vis lis-keys)
    (del-key :gen-pla lis-keys)
    (del-key :dev-vis lis-keys)

    ; Llamamos a extraer funciones de agentes, ojo,
    ; con lista invertida y partida-internal da la vuelta
    (partida-internal2 lis-nom-agentes lis-keys))

(defun partition-args (keywords arglist) 
;; Thanks to Barry Margolin:
;; Subject: Re: &rest and &key;  group: comp.lang.lisp:
;; http://groups.google.com/group/comp.lang.lisp/tree/browse_frm/thread/dcf71f6bdd232985/a9c3e3362c141ae2?rnum=1&q=%26rest+%26key+defun&_done=%2Fgroup%2Fcomp.lang.lisp%2Fbrowse_frm%2Fthread%2Fdcf71f6bdd232985%2Fa9c3e3362c141ae2%3Ftvc%3D1%26q%3D%26rest+%26key+defun%26#doc_4a9a5c342e93a827
  "Partition the argument list around the first argument that is in KEYWORDS." 
  (let ((first-key (position-if #'(lambda (elt) (member elt keywords)) 
                                arglist)))
    (values (subseq arglist 0 first-key)
            (and first-key (subseq arglist first-key)))))

(defun partida (&rest arglist)
    "Inicia partida del juego PlanTron entre agentes pasados. Para parámetros reales ver función PARTIDA-INTERNAL. Se le ha de pasar los nombres de los agentes a enfrentarse (ver PT:JUG). Ademas acepta varia keywords: La key VIS indica el número del agente de visualización AGENTE-PT-VIS que se  utilizará (1) o si no se utiliza (nil). GEN-PLA indica el número o lista de números de agentes generadores de plantas a utilizar. DEV-VIS indica un dispositivo normalmente de terminal (/dev/pts/*) donde emitir la salida del agente visualizador seleccionado. El resto de keywords definen el tipo de partida a jugar (ver PT-BASE::EXTRC-TIPO-PARTIDA)."
  (multiple-value-bind (lis-nom-agentes lis-keys) 
      ; Se pone en LIS-NOM-AGENTES la lista de nombres de agentes pasados
      ; y en LIS-KEYS la lista de keywords
      (partition-args keywords-partida arglist)
    (apply #'partida-internal lis-nom-agentes lis-keys)))


(defun partida-todos-internal (args &rest lis-keys &key (num 1) &allow-other-keys)
  "Compite con agentes en memoria, emitiendo una lista final con sublistas conteniendo: nombre del agente, número de pasos efecutados, energía final acumulada y número de partidas jugadas."
  (let ((lis-jug (if (first args)
		     (first args)
		   (borra-ag-interactivo (jug))))
	(lis-res (second args)))
    ; Lista asociativa de resultados
    ;     cada uno con: (paso_dados enegia_final_ac num_partidas) 
    (dolist (ag lis-jug)
      (if (null (busca-ag ag lis-res))
          (push (list ag 0 0 0) lis-res)))
    ;Lanzamos partidas
    (del-key :num lis-keys)
    (dotimes (i num)
      (dolist (ag lis-jug)
	(let* ((res (apply #'partida-internal (list ag) lis-keys))
	       (res-ag (assoc id-ag-eval (second res)))
	       (stad (busca-ag ag lis-res)))
	  (setf (first stad)  (+ (first stad)  (second res-ag)))
	  (setf (second stad) (+ (second stad) (third  res-ag)))
	  (setf (third stad)  (1+ (third stad))))))
    (sort lis-res #'>
	  :key #'(lambda (s)
		   (/ (+ (second s) (third s)) (fourth s))))))

(defun partida-todos (&rest arglist)
  "Compite todos los agentes en memoria, emitiendo una lista final con sublistas conteniendo: nombre del agente, número de pasos efecutados, energía final acumulada y número de partidas jugadas. El resultado final es ordenado por número de pasos efecutados mas la engergía final acumulada y todo dividido entre número de partidas jugadas. Los parámetros que se le pasan realmente son (&OPTIONAL LIS-AG LIS-RES &KEY PT-KEYS), donde LIS-AG es una lista con los nombres agentes a jugar, si no se pasa se inicializan con el resultado de (JUG) una vez eliminado los agentes interactivos. LIS-RES es una lista de resultados anteriores, a la que se añadirán los resultados de las partidas que se efectúen. PT-KEYS son las keywords a pasar a la función PARTIDA-INTERNAL."
  (multiple-value-bind (args lis-keys) 
      ; Se pone en ARGS los argumentos pasados
      ; y en LIS-KEYS la lista de keywords
      (partition-args (cons :num keywords-partida) arglist)
    (apply #'partida-todos-internal args lis-keys)))

(defun enfrenta2 (ag1 ag2 arglist lis-res) ; FIXME
  "Enfrenta a dos agentes PlanTron."
  (let* ((pareja (list ag1 ag2))
	 (res (apply #'partida-internal pareja arglist))
	 (stad1 (busca-ag ag1 lis-res))
	 (stad2 (busca-ag ag2 lis-res)))
    (setf (second stad1) (1+ (second stad1))
	  (second stad2) (1+ (second stad2)))
    (dolist (ag-pierde (first res))
      (let ((stad-per (busca-ag (nth (1- ag-pierde) pareja) lis-res)))
	(setf (first stad-per) (1+ (first stad-per)))))))


(defun partida-todos2-internal (args &rest lis-keys &key (num 1) &allow-other-keys)
  "Compite agentes en memoria todos contra todos dos a dos, emitiendo una lista final con sublistas conteniendo: nombre del agente, número de partidas perdidas y número de partidas jugadas. El resultado final es ordenado por número de partidas perdidas dividido entre número de partidas jugadas."
  (let* ((lis-jug-1 (if (first args) (borra-ag-interactivo (first args))
			(borra-ag-interactivo (jug))))
	 (lis-jug-2 (if (second args) (borra-ag-interactivo (second args))
			(borra-ag-interactivo (jug))))
	 (lis-res (third args)))
         ; Lista asociativa de resultados
    (dolist (ag (append lis-jug-1 lis-jug-2))
      (if (null (busca-ag ag lis-res))
	  (push (list ag 0 0) lis-res)))
    ; Enfrentamientos restringidos
    (del-key :num lis-keys)
    (dolist (ag1 lis-jug-1)
      (dolist (ag2 lis-jug-2) ; (progn (print lis-res)
				;    (force-output)))
	(dotimes (i num)
	  (if (string/= ag1 ag2)
	      (enfrenta2 ag1 ag2 lis-keys lis-res)))))
    (dolist (ag lis-jug-1)
      (setf lis-jug-2 (remove ag lis-jug-2 :test #'string=)))
    (dolist (ag1 lis-jug-2)
      (dolist (ag2 lis-jug-1) ; (progn (print lis-res)
				;    (force-output)))
	(dotimes (i num)
	  (if (string/= ag1 ag2)
	      (enfrenta2 ag1 ag2 lis-keys lis-res)))))
    (sort lis-res #'<
	  :key #'(lambda (s)
		   (/ (second s) (third s))))))


(defun partida-todos2 (&rest arglist)
  "Compite agentes en memoria todos contra todos dos a dos, emitiendo una lista final con sublistas conteniendo: nombre del agente, número de partidas perdidas y número de partidas jugadas. El resultado final es ordenado por número de pasos efecutados mas la engergía final acumulada y todo dividido entre número de partidas jugadas. Los parámetros que se le pasan realmente son (&OPTIONAL LIST-AG-1 LIST-AG-2 LIS-RES &KEY PT-KEYS), donde LIST-AG-1 y LIST-AG-2 son listas de nombres agentes a enfrentar entre sí, si no se pasa se inicializan con el resultado de (JUG) una vez eliminado los agentes interactivos. LIS-RES es una lista de resultados anteriores, a la que se añadirán los resultados de las partidas que se efectúen. PT-KEYS son las keywords a pasar a la función PT-INTERNAL."
  (multiple-value-bind (args lis-keys) 
      ; Se pone en ARGS los argumentos pasados
      ; y en LIS-KEYS la lista de keywords
      (partition-args (cons :num keywords-partida) arglist)
    (apply #'partida-todos2-internal args lis-keys)))


;;
;; Inicialización
;; --------------

; Extraemos la versión del entorno del fichero VERSION
(let ((st  (open (merge-pathnames (pathname "VERSION") path) 
	    :direction :input :if-does-not-exist :error)))
  (setf pt-base::version-entorno (read st))
  (close st))

; Inicializamos directorio de mapas de tableros
(setf pt-base::*path-maps* (if (directory (merge-pathnames (pathname "maps") path))
			       (merge-pathnames (pathname "maps/") path)))

; Cargamos agentes visualizadores
(let ((lis-agentes-tmp lis-agentes))
  (setf lis-agentes nil)
  (load-ag-1 (merge-pathnames (pathname "pt-vis") path))
  (setf lis-ag-vis (reverse lis-agentes))
  (setf lis-agentes (append lis-ag-vis lis-agentes-tmp)))

;; (load (concatenate 'string path-trho "pt-log"))
;;(setf lis-ag-log (rest (first lis-agentes)))
;;(setf lis-agentes nil)

; Cargamos agentes generadores de plantas
(let ((lis-agentes-tmp lis-agentes))
  (setf lis-agentes nil)
  (load-ag-1 (merge-pathnames (pathname "pt-agentes-genplc") path))
  (setf lis-ag-genpla (reverse lis-agentes))
  (setf lis-agentes (append lis-ag-genpla lis-agentes-tmp)))

;; Cargamos agentes de utilidades
(load-ag-1 (merge-pathnames (pathname "pt-util") path))

; Cargamos agentes generadores de direcciones
(let ((lis-agentes-tmp lis-agentes))
  (setf lis-agentes nil)
  (load-ag-1 (merge-pathnames (pathname "pt-agentes-gendir") path))
  ;; (load-ag-1 (merge-pathnames (pathname "dpdd-minmax") path))
  (setf lis-ag-gendir (reverse lis-agentes))
  (setf lis-agentes (append lis-ag-gendir lis-agentes-tmp)))

;; Inicializamos variables de acuerdo a juego deseado
(progn
  (setf pt-base:*ratio-ene-jug-ini* 0.5)
  (setf pt-base:*mov-con* t)
  (setf pt-base:*tam-col* t)
  (setf pt-base:*pos-alea* nil)
  (setf pt-base:*for-cab* 3)
  (setf pt-base:*for-col* 1)
  (setf *lis-nom-agentes-def* '("HUMANO"))
  (setf *gen-pla* 1)
  ; Por si ha cargado dpdd-minmax
  (setf *gen-dir* (if (>= (length lis-ag-gendir) 5) 4 1))
;;  (setf *gen-dir* nil)
  t)

